/*
 * taskview-application-proxy.c
 *
 * TaskviewApplicationProxy
 *
 * Copyright © 2010 Salomon Sickert
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the licence or (at
 * your option) any later version.
 *
 * See the included COPYING file for more information.
 *
 * Authors:
 *    Salomon Sickert <sickert@in.tum.de>
 */

#include "taskview.h"

#define TASKVIEW_APPLICATION_PROXY_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), TASKVIEW_TYPE_APPLICATION_PROXY, TaskviewApplicationProxyPrivate))

G_DEFINE_TYPE (TaskviewApplicationProxy, taskview_application_proxy, G_TYPE_OBJECT);

struct _TaskviewApplicationProxyPrivate
{
  GDBusConnection *connection;
  GDBusProxy *manager;
  gchar *dbus_name;
  GHashTable *view_proxies;
};

enum /* signals */
{
  PROPERTY_CHANGED,
  TASKVIEW_ADDED,
  TASKVIEW_REMOVED,
  LAST_SIGNAL
};

static guint _signals[LAST_SIGNAL] = { 0, 0, 0, };

/* ----------------------------------------------------------------------------------------------------
 * GObject Properties
 * ---------------------------------------------------------------------------------------------------- */

enum /* properties */
{
  PROP_0,
  PROP_DBUS_NAME,
  PROP_DBUS_CONNECTION,
  PROP_APPLICATION_NAME
};

static void
taskview_application_proxy_set_property (GObject      *object,
    guint         property_id,
    const GValue *value,
    GParamSpec   *pspec)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (object);

  switch (property_id)
    {
    case PROP_DBUS_NAME:
      /* Nothing to free, as it is G_PARAM_CONSTRUCT_ONLY */
      self->priv->dbus_name = g_value_dup_string (value);
      break;

    case PROP_DBUS_CONNECTION:
      /* Nothing to free, as it is G_PARAM_CONSTRUCT_ONLY */
      self->priv->connection = G_DBUS_CONNECTION (g_value_get_pointer (value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
taskview_application_proxy_get_property (GObject    *object,
    guint       property_id,
    GValue     *value,
    GParamSpec *pspec)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (object);
  GVariant *app_name;

  switch (property_id)
    {
    case PROP_DBUS_NAME:
      g_value_set_string (value, self->priv->dbus_name);
      break;

    case PROP_APPLICATION_NAME:
      g_value_set_string (value, "");
      if (self->priv->manager != NULL)
        {
          app_name = g_dbus_proxy_get_cached_property (self->priv->manager, "ApplicationName");

          if (app_name != NULL)
            {
              g_value_set_string (value, g_variant_get_string (app_name, NULL));
              g_variant_unref (app_name);
            }
        }
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/* ----------------------------------------------------------------------------------------------------
 * Helper
 * ---------------------------------------------------------------------------------------------------- */

static void
create_proxy (TaskviewApplicationProxy *self, const gchar *path)
{
  GObject *proxy;
  GError *error;
  GVariant *result;
  GDBusNodeInfo *node;
  const gchar *xml_data;
  guint n;
  GType type;

	/* FIXME: Make async */
	
  error = NULL;
  result = g_dbus_connection_call_sync (self->priv->connection,
                                        self->priv->dbus_name,
                                        path,
                                        "org.freedesktop.DBus.Introspectable",
                                        "Introspect",
                                        NULL,
                                        G_VARIANT_TYPE ("(s)"),
                                        G_DBUS_CALL_FLAGS_NONE,
                                        3000, /* 3 sec */
                                        NULL,
                                        &error);

	if (result == NULL)
    {
      g_printerr ("Error: %s\n", error->message);
      g_error_free (error);
      goto out;
    }

  g_variant_get (result, "(&s)", &xml_data);
  node = g_dbus_node_info_new_for_xml (xml_data, &error);

	if (node == NULL)
    {
      g_printerr ("Error parsing introspection XML: %s\n", error->message);
      g_error_free (error);
      goto out;
    }

	// We search for an appropiate interface, falling back to generic...
	type = TASKVIEW_TYPE_GENERIC_PROXY;
	
  for (n = 0; node->interfaces != NULL && node->interfaces[n] != NULL; n++)
    if (g_strcmp0 (node->interfaces[n]->name, TASKVIEW_IO_INTERFACE_NAME) == 0)
      type = TASKVIEW_TYPE_IO_PROXY;
        
  proxy = g_object_new (type,
                        "dbus-name", self->priv->dbus_name,
                        "dbus-object-path", path,
                        "dbus-connection", g_object_ref (self->priv->connection),
                        NULL);

  g_hash_table_insert (self->priv->view_proxies, g_strdup (path), proxy);
  g_signal_emit_by_name (self, "taskview-added", path);
  g_debug ("Created %s / Table: %i", path, g_hash_table_size (self->priv->view_proxies));

out:
  if (node != NULL)
    g_dbus_node_info_unref (node);
  if (result != NULL)
    g_variant_unref (result);
}

static void
remove_proxy (TaskviewApplicationProxy *self, const gchar *path)
{
  g_hash_table_remove (self->priv->view_proxies, path);
  g_signal_emit_by_name (self, "taskview-removed", path);
  g_debug ("Removed %s / Table: %i", path, g_hash_table_size (self->priv->view_proxies));
}

/* ----------------------------------------------------------------------------------------------------
 * Signal Handler
 * ---------------------------------------------------------------------------------------------------- */

static void
on_properties_changed (GDBusProxy          *proxy,
                       GVariant            *changed_properties,
                       const gchar* const  *invalidated_properties,
                       gpointer             user_data)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (user_data);
  g_signal_emit_by_name (self, "properties-changed");
}

static void
on_signal (GDBusProxy *proxy,
           gchar      *sender_name,
           gchar      *signal_name,
           GVariant   *parameters,
           gpointer    user_data)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (user_data);
  GVariant *unboxed;
  gchar *parameters_str;

  if (g_strcmp0 (signal_name, "TaskViewAdded") == 0)
    {
      unboxed = g_variant_get_child_value (parameters, 0);
      parameters_str = g_variant_dup_string (unboxed, NULL);
      create_proxy (self, parameters_str);
    }
  else if (g_strcmp0 (signal_name, "TaskViewRemoved") == 0)
    {
      unboxed = g_variant_get_child_value (parameters, 0);
      parameters_str = g_variant_dup_string (unboxed, NULL);
      remove_proxy (self, parameters_str);
    }
  else
    {
      parameters_str = g_variant_print (parameters, TRUE);
      g_debug ("Received Unkown Signal: %s: %s\n",
               signal_name,
               parameters_str);
    }

  g_free (parameters_str);
  g_variant_unref (unboxed);
}


static void
on_proxy_appeared (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
  /* FIXME: Remove me, after testing */
  g_assert (TASKVIEW_IS_APPLICATION_PROXY (user_data));
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (user_data);
  GError *error;
  GDBusProxy *proxy;
  GVariant *result;
  GVariant *sarray;
  GVariant *path;
  gint len;
  gint pos;

  error = NULL;
  proxy =  g_dbus_proxy_new_for_bus_finish (res, &error);

  if (proxy == NULL)
    {
      g_printerr ("Error creating proxy: %s\n", error->message);
      g_error_free (error);
      return;
    }

  g_signal_connect (proxy,
                    "g-properties-changed",
                    G_CALLBACK (on_properties_changed),
                    self);

  g_signal_connect (proxy,
                    "g-signal",
                    G_CALLBACK (on_signal),
                    self);


  self->priv->manager = proxy;

  /*FIXME: ASYNC */
  result = g_dbus_proxy_call_sync (proxy,
                                   "TaskViewList",
                                   NULL,
                                   G_DBUS_CALL_FLAGS_NONE,
                                   -1,
                                   NULL,
                                   &error);
	if (error != NULL)
    {
      g_error ("%s", error->message);
      g_error_free (error);
			// TODO: Fallback!
			return;
    }

  sarray = g_variant_get_child_value (result, 0);
  len = g_variant_n_children (sarray);

  /* FIXME: Replace with iterator */
  for (pos = 0; pos < len; pos++)
    {
      path = g_variant_get_child_value (sarray, pos);
      create_proxy (self, g_variant_get_string (path, NULL));
      g_variant_unref (path);
    }

  g_variant_unref (sarray);
  g_variant_unref (result);
  g_signal_emit_by_name (self, "properties-changed");
}

/* ----------------------------------------------------------------------------------------------------
 * GObject
 * ---------------------------------------------------------------------------------------------------- */

static void
taskview_application_proxy_constructed (GObject *gobject)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (gobject);

  g_assert (g_dbus_is_name (self->priv->dbus_name));
  g_assert (G_IS_DBUS_CONNECTION (self->priv->connection));

  /* FIXME: How do we know, if the proxy vanished? */
  g_dbus_proxy_new (self->priv->connection,
                    G_BUS_NAME_WATCHER_FLAGS_NONE,
                    NULL, /* FIXME: Add org.gnome.TaskView.Manager */
                    self->priv->dbus_name,
                    TASKVIEW_MANAGER_OBJECTPATH,
                    TASKVIEW_MANAGER_INTERFACE_NAME,
                    NULL,
                    on_proxy_appeared,
                    self);
}

static void
taskview_application_proxy_dispose (GObject *gobject)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (gobject);

  if (self->priv->view_proxies)
    {
      g_hash_table_unref (self->priv->view_proxies);
      self->priv->view_proxies = NULL;
    }

  if (self->priv->manager != NULL)
    {
      g_object_unref (self->priv->manager);
      self->priv->manager = NULL;
    }

  if (self->priv->connection != NULL)
    {
      g_object_unref (self->priv->connection);
      self->priv->connection = NULL;
    }

  G_OBJECT_CLASS (taskview_application_proxy_parent_class)->dispose (gobject);
}

static void
taskview_application_proxy_finalize (GObject *gobject)
{
  TaskviewApplicationProxy *self = TASKVIEW_APPLICATION_PROXY (gobject);

  g_free (self->priv->dbus_name);

  G_OBJECT_CLASS (taskview_application_proxy_parent_class)->finalize (gobject);
}

static void
taskview_application_proxy_class_init (TaskviewApplicationProxyClass *klass)
{
  GParamSpec *pspec;
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (gobject_class, sizeof (TaskviewApplicationProxyPrivate));

  _signals[TASKVIEW_ADDED] = g_signal_new ("taskview-added",
                             TASKVIEW_TYPE_APPLICATION_PROXY,
                             G_SIGNAL_RUN_LAST,
                             0,
                             NULL,
                             NULL,
                             g_cclosure_marshal_VOID__STRING,
                             G_TYPE_NONE /* return_type */,
                             1     /* n_params */,
                             G_TYPE_STRING  /* param_types */);


  _signals[TASKVIEW_REMOVED] = g_signal_new ("taskview-removed",
                               TASKVIEW_TYPE_APPLICATION_PROXY,
                               G_SIGNAL_RUN_LAST,
                               0,
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__STRING,
                               G_TYPE_NONE /* return_type */,
                               1     /* n_params */,
                               G_TYPE_STRING  /* param_types */);

  _signals[PROPERTY_CHANGED] = g_signal_new ("properties-changed",
                               TASKVIEW_TYPE_APPLICATION_PROXY,
                               G_SIGNAL_RUN_LAST,
                               0,
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE /* return_type */,
                               0     /* n_params */,
                               NULL  /* param_types */);

  gobject_class->constructed = taskview_application_proxy_constructed;
  gobject_class->dispose = taskview_application_proxy_dispose;
  gobject_class->finalize = taskview_application_proxy_finalize;
  gobject_class->set_property = taskview_application_proxy_set_property;
  gobject_class->get_property = taskview_application_proxy_get_property;

  /* D-Bus Name */
  pspec = g_param_spec_string ("dbus-name",
                               "D-Bus Name",
                               "The D-Bus name of the remote application to monitor",
                               NULL,
                               G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
  g_object_class_install_property (gobject_class, PROP_DBUS_NAME, pspec);

  /* D-Bus Connection */
  pspec = g_param_spec_pointer ("dbus-connection",
                                "D-Bus connection",
                                "The D-Bus connection",
                                G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
  g_object_class_install_property (gobject_class, PROP_DBUS_CONNECTION, pspec);

  /* Application Name */
  pspec = g_param_spec_string ("application-name",
                               "Application Name",
                               "The name of the remote application",
                               "" /* default value */,
                               G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
  g_object_class_install_property (gobject_class, PROP_APPLICATION_NAME, pspec);
}

static void
taskview_application_proxy_init (TaskviewApplicationProxy *self)
{
  self->priv = TASKVIEW_APPLICATION_PROXY_GET_PRIVATE (self);
  self->priv->view_proxies = g_hash_table_new_full (g_str_hash,
                             g_str_equal,
                             g_free,
                             g_object_unref);
}

const GHashTable *
taskview_application_proxy_get_views (TaskviewApplicationProxy *proxy)
{
  g_return_val_if_fail (TASKVIEW_IS_APPLICATION_PROXY (proxy), NULL);
  return proxy->priv->view_proxies;
}

TaskviewIOProxy *  
taskview_application_proxy_get_io_proxy (TaskviewApplicationProxy *proxy, const gchar *object_path)
{
  gpointer data;
  
  g_return_val_if_fail (TASKVIEW_IS_APPLICATION_PROXY (proxy), NULL);

  data = g_hash_table_lookup (proxy->priv->view_proxies, object_path);

  if (!TASKVIEW_IS_IO_PROXY (data))
    return NULL;
  
  return TASKVIEW_IO_PROXY (data);
}

TaskviewGenericProxy * taskview_application_proxy_get_generic_proxy (TaskviewApplicationProxy *proxy, const gchar *object_path)
{
  gpointer data;
  
  g_return_val_if_fail (TASKVIEW_IS_APPLICATION_PROXY (proxy), NULL);

  data = g_hash_table_lookup (proxy->priv->view_proxies, object_path);

  if (!TASKVIEW_IS_GENERIC_PROXY (data))
    return NULL;
  
  return TASKVIEW_GENERIC_PROXY (data);
}

/* ----------------------------------------------------------------------------------------------------
 * C Wrapper for g_object_get
 * ---------------------------------------------------------------------------------------------------- */

gchar * 
taskview_application_proxy_get_application_name (TaskviewApplicationProxy *self)
{
  gchar *value;
  g_object_get (self, "application-name", &value, NULL);
  return value;
}
